#include <string>
#include <iostream>
#include <list>
#include <vector>


#include "ipaddress.h"
#include "router.h"
#include "packet.h"

 
#include <string.h>

#define PKT_BUFF_SIZE 100

Router::Router(string name)
{
  name_ = name;
  interfaceCount_ = 0;
  dataLinks_ = new DataLink *[RTR_MAXDATALINKS];
  IPAddresses_ = new IPAddress *[RTR_MAXDATALINKS];

 
  routeCount_ = 0;

  
}

Router::~Router()
{
  ;
}

/* this routine is called by the configreader to add another
   interface to a datalink to the router */
void Router::addInterface(DataLink * network, IPAddress address)
{

  dataLinks_[interfaceCount_] = network;
  IPAddresses_[interfaceCount_] = new IPAddress(address);
  interfaceCount_++;

  /* resize the number of outbound packet queues */
  sendQ.resize(interfaceCount_);
  linkQuotas.resize(interfaceCount_);
  
/* let the datalink know we're attached to it */
  network->addRouter(this, address);
 
}

/* Output information about interfaces on standard out */
void Router::printInterfaces()
{

  cout << "---- Interface Summary on " << name_ << " ----" << endl;
  for(unsigned int i = 0; i < interfaceCount_; i++)
    {
      cout << "if" << i << ": " << IPAddresses_[i]->toString() << endl;
    }
}

/* return the ID of the outbound interface for a packet;  if the packet's
   destination is not on a directly connected network, return -1  */
int Router::destInterface(IPAddress a)
{
  for(unsigned int i =0; i < interfaceCount_; i++)
    if( IPAddresses_[i]->match(a))
      return i;
  
    return -1;
}

void Router::deliverPacket(Packet *p)
{
  recvQueue.push_back(p);
}


/* Place a packet in the proper outbound queue */
void Router::queuePacket(Packet *p)
{
  Qentry qe;
  
  int interface;
  IPAddress nxHop;
  
  if(forMe(p))
    {
      localQ.push(p);
      return;
    }
  
  /* forward packet */
  
  // is the packet destined for a directly connected network? 
  int localI = localInterface(p->dst());
  if(localI != -1)
    {
      interface = localI;
      nxHop = p->dst();
    }
  else
    {
      nxHop = nextHop(p->dst());
      if(nxHop != IPAddress("0.0.0.0/32"))
	interface = localInterface(nxHop);
      else
	{
	  cout << name_ << ": Can't reach " << p->dst().toStringPrefix() << endl;
	  return;
	}
    }
  
  qe.p = p;
  qe.nxHop = nxHop;

  /* put the packet in the proper outbound queue */

  if(sendQ[interface].size() < PKT_BUFF_SIZE)
    sendQ[interface].push_back(qe);

}


/* move packets from the general receive queue to 
   specific outbound queues */
void Router::rippleQueue()
{
  if(recvQueue.size())
    cout << name_ << ": Inbound "  << recvQueue.size() << " packets." << endl;
  for(unsigned int i =0; i < 3*  recvQueue.size(); i ++)
    {
      int x = (int) (rand()) % recvQueue.size();
      int y = (int) (rand()) % recvQueue.size();
     
      //cout << "swapping " << x << " and " << y << endl;
      /* Swap the packets in slots x and y */
      Packet * t;
      t = recvQueue[x];
      recvQueue[x] = recvQueue[y];
      recvQueue[y] = t;
      
    }

  for(unsigned int i = 0; i < recvQueue.size(); i++)
    {
      queuePacket(recvQueue[i]);
    }
  recvQueue.clear();
}

void Router::runTimeSlice(int timeWidth)
{
  //  cout << name_ << ": Running for " << timeWidth << endl;
 
  while(!localQ.empty())
    {
      handleLocalPacket(localQ.front());
      localQ.pop();
    }

  // linkquotas now are a member variable
  // vector<int> linkQuotas;

  int maxQuota = 0;
  for(unsigned int i =0; i < linkQuotas.size(); i++)
    {
      linkQuotas[i] = int(dataLinks_[i]->bandwidth() * timeWidth * .001 / 8);
      if(linkQuotas[i] > maxQuota) maxQuota = linkQuotas[i];
    }

  
    // display linkQuotas
  /*
    cout << name_ << ": linkQuotas: ";
    for(unsigned int i =0; i < linkQuotas.size(); i ++)
    cout << linkQuotas[i] << ", ";
    cout << endl;
  */ 
  
  for(unsigned int iface  =0; iface < interfaceCount_; iface++)
    {
      bool quotaLeft = true; // this gets set to false when we run out of linkQuota
      while(quotaLeft && !sendQ[iface].empty())
	{
	  Qentry qe = sendQ[iface].front();
	  
	  int packetCost = qe.p->tsize();
	  //cout << "pktCost : " << packetCost << endl;
	  if(packetCost < linkQuotas[iface])  // enough linkquota left to send the pkt?
	    {
	      if(servicePacket(qe.p))  // do maintanence like check TTL, record route, etc.
		{
		  qe.p->setTsize(qe.p->size());
		  dataLinks_[iface]->deliverPacket(qe.p, qe.nxHop); // deliver packet to datalink
		  sendQ[iface].pop_front(); // remove packet from queue
		  linkQuotas[iface] -= packetCost; // decrease linkquota as appropriate
		}
	    }
	  else
	    {
	      /* decrease the temp size of the packet; e.g., start sending it */
	      qe.p->setTsize(packetCost - linkQuotas[iface]);
	      linkQuotas[iface] = 0;
	      quotaLeft=false;
	      /* Delay packets in queue that weren't serviced yet. */
	      delayQueue(sendQ[iface], timeWidth);
	    }
	}
    }
}

/* add some delay to all packets in an interface queue */
void Router::delayQueue(list<Qentry> sendQueue, int timeWidth)
{
  
  cout << name_ << " delaying " << sendQueue.size() << " packets." << endl;
  for(list<Qentry>::const_iterator itor=sendQueue.begin(); itor !=sendQueue.end(); itor ++)
    {
      (*itor).p->setDelay((*itor).p->delay()+timeWidth);
    }
}

/* a packet arrived for this host.  Do whatever with it */
void Router::handleLocalPacket(Packet *p)
{
  // cout << name_ << ": Pkt from " << p->src().toString() << " Delay: ";
  // cout << p->delay() << " TTL: " << p->TTL() << " Payload: " << p->payload() << endl;

  p->deliver(); // set the packet delivered flag to true

}

/* A packet is getting delivered.  Do necessary stuff like decrement TTL, 
   record route if necessary, etc.  Return 1 if everything is OK, return
   0 if something went wrong and the packet must be dropped.
*/

int  Router::servicePacket(Packet *p)
{
  if(p-> recordRoute())
    p->appendPayload(name_ + '\n');
  p->setTTL(p->TTL()-1); 
  if(p->TTL() ==0) /* drop packets whoose TTL's have expired */
    {
      cout << name_ << ": TTL expired in transit." << endl;
      return 0;
    }
  return 1;
}

bool Router::forMe(Packet *p)
{
  for(unsigned int i = 0; i < interfaceCount_; i++)
    {
      if(*IPAddresses_[i] == p->dst() || (p->dst().isBroadcast() && (p->dst().net()==IPAddresses_[i]->net()) ))
	return true;
    }
  return false;
}

/* Return the ID of the local interface connected to the network of the 
   given address' destination.  Return -1 if it cannot be serviced by
   a directly connected network
*/

int Router::localInterface(IPAddress a)
{
  for(unsigned int i = 0; i < interfaceCount_; i++)
    {
      if(IPAddresses_[i]->match(a))
	return i;
    }
  return -1;
}

/* Returns the next hop address for an address.  Assumes that 
   the address is not on a directly connected network. 

   Return 0.0.0.0/32 if the address is unreachable.

*/
IPAddress Router::nextHop(IPAddress a)
{
  int i =0;

  while(i < routeCount_)
    {

      if(routeTable_[i][0].match(a))
	return routeTable_[i][1];
      i++;
    }
  return IPAddress("0.0.0.0/32"); 
}

/* Add a route to the routing table.  

     !! Replace all this when new routing table is landed !!
*/

void Router::addRoute(IPAddress dst, IPAddress nextHop, int metric)
{
  routeTable_[routeCount_][0] = dst;
  routeTable_[routeCount_][1] = nextHop;
  routeCount_++;
  
  metric++; // suppress compiler warning, we ignore metrics now
}


/*
  Print a copy of the routing table to standard out 
*/
void Router::printRoutes()
{
  cout << "  -- Routes on " << name_  << " -- " << endl;
  cout << "Destination          Next Hop " << endl;
  cout << "------------------------------" << endl;
  
  for(int i =0; i < routeCount_; i++)
    {
      cout << routeTable_[i][0].toStringPrefix() << "          " << routeTable_[i][1].toString() << endl;
    }
}








